package insilico.core.descriptor.nestle;

import insilico.core.descriptor.Descriptor;
import insilico.core.descriptor.DescriptorBlock;
import insilico.core.descriptor.blocks.AtomCenteredFragments;
import insilico.core.descriptor.blocks.Constitutional;
import insilico.core.descriptor.blocks.FunctionalGroups;
import insilico.core.descriptor.blocks.logp.MLogP;
import insilico.core.exception.DescriptorNotFoundException;
import insilico.core.molecule.InsilicoMolecule;
import insilico.core.molecule.conversion.file.MoleculeFileSmiles;
import java.io.PrintWriter;
import java.util.ArrayList;

/**
 * Descriptors needed by Nestle' models, based on VEGA core descriptor with
 * some modification in order to be compliant with older Dragon descriptors
 * used in their original model.
 * 
 * @author Alberto Manganaro (a.manganaro@kode-solutions.net)
 */
public class DescriptorsNestle {

    
    public DescriptorsNestle() {
        
    }
    
    
    /**
     * Main method to calculate the descriptors, from a given input file 
     * containing just the SMILES of the molecules. The results are written
     * in a tab-separated text file.
     * 
     * @param InputFile full path+filename of the input file (with the SMILES
     * to be processed)
     * @param OutFile full path+filename of the file to be written with the
     * calculated descriptors
     */
    public void Calculate(String InputFile, String OutFile) {
        
        ArrayList<InsilicoMolecule> mols;
        
        MoleculeFileSmiles sfile = new MoleculeFileSmiles();
        try {
            sfile.OpenFile(InputFile);
            mols = sfile.ReadAll();        
        } catch (Throwable ex) {
            System.out.println("Unable to read and import molecules - " + ex.getMessage());
            return;
        }
        
        PrintWriter wr;
        try {
            wr = new PrintWriter(OutFile);
        } catch (Throwable ex) {
            System.out.println("Unable to open dest file - " + ex.getMessage());
            return;
        }
        
        // header
        String head = "no.\tSMILES";
        head += "\t" + "RBN";
        head += "\t" + "nN";
        head += "\t" + "nHM";
        head += "\t" + "nX";
        head += "\t" + "nCt";
        head += "\t" + "nArCOOH";
        head += "\t" + "nOHp";
        head += "\t" + "nArOR";
        head += "\t" + "nRSR";
        head += "\t" + "nSO";
        head += "\t" + "nCXr=";
        head += "\t" + "nCconjX";
        head += "\t" + "H-052";
        head += "\t" + "H-054";
        head += "\t" + "O-058";
        head += "\t" + "S-107";
        head += "\t" + "P-117";
        head += "\t" + "Hy";
        head += "\t" + "MLOGP";
        head += "\t" + "MLOGP2";
        wr.write(head + System.lineSeparator());
        wr.flush();
                                
        // build descriptors blocks
        AtomCenteredFragments b_acf = new AtomCenteredFragments();
        Constitutional b_const = new Constitutional();
        FunctionalGroups b_func = new FunctionalGroups();
        MLogP b_mlogp = new MLogP();
        
        // run calculation
        int idx = 0;
        int dNum = 0;
        int nErr = 0;
        for (InsilicoMolecule m  : mols) {
            
            idx++;
            System.out.println("* Processing mol no. " + idx);
            
            // check mol
            if (!m.IsValid()) {
                nErr++;
                System.out.println("INVALID MOLECULE\t" + idx);
                String s = idx + "\tERROR";
                for (int i=0; i<20; i++)
                    s += "\t" + Descriptor.MISSING_VALUE_STR;
                s += System.lineSeparator();
                wr.write(s);
                wr.flush();
                continue;
            }
            
            // calculate
            b_mlogp.Calculate(m);
            b_const.Calculate(m);
            b_func.Calculate(m);
            b_acf.Calculate(m);
            CustomDescriptors otherDesc = new CustomDescriptors(m);
            
            double MLogP;
            try {
                MLogP = b_mlogp.GetByName("MLogP").getValue();
                MLogP = MLogP + otherDesc.N_MlogP_Correction * (-3.684);
            } catch (DescriptorNotFoundException ex) {
                MLogP = -999;
            }
            
            double MLogP_square = -999;
            if (MLogP != -999)
                MLogP_square = Math.pow(MLogP,2);
            
                        
            // create output
            String s = idx + "\t" + m.GetSMILES();
            
            s += "\t" + otherDesc.RBN;
            s += "\t" + getDescriptor(b_const, "nN");
            s += "\t" + otherDesc.nHM;
            s += "\t" + getDescriptor(b_const, "nX");
            s += "\t" + getDescriptor(b_func, "nCt");
            s += "\t" + getDescriptor(b_func, "nArCOOH");
            s += "\t" + getDescriptor(b_func, "nOHp");
            s += "\t" + getDescriptor(b_func, "nArOR");
            s += "\t" + getDescriptor(b_func, "nRSR");
            s += "\t" + getDescriptor(b_func, "nSO");
            s += "\t" + getDescriptor(b_func, "nCXr=");
            s += "\t" + getDescriptor(b_func, "nCconjX");
            s += "\t" + getDescriptor(b_acf, "H-052");
            s += "\t" + getDescriptor(b_acf, "H-054");
            s += "\t" + getDescriptor(b_acf, "O-058");
            s += "\t" + getDescriptor(b_acf, "S-107");
            s += "\t" + getDescriptor(b_acf, "P-117");
            s += "\t" + otherDesc.Hy;
            s += "\t" + MLogP;
            s += "\t" + MLogP_square;
            
            // write
            wr.write(s + System.lineSeparator());
            wr.flush();
        }
        
        wr.close();
        System.out.println();
        System.out.println("* Finished, total descriptors calculated: " + dNum);
        System.out.println("* Processed molecules: " + idx);
        System.out.println("* Invalid molecules: " + nErr);
    }


    private String getDescriptor(DescriptorBlock block, String name) {
        String res;
        try {
            res = block.GetByName(name).getFormattedValue();
        } catch (DescriptorNotFoundException e) {
            return "-999";
        }
        return res;
    }
    
    
    
    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        
        // just for local testing!        
        DescriptorsNestle dd = new DescriptorsNestle();
        dd.Calculate("D:\\Dropbox (kode)\\tech_manganaro\\IRFMN\\In progress\\Helma - aggiornamento modelli Nestlè\\descrittori D5 completi\\helma.smi", 
                     "D:\\Dropbox (kode)\\tech_manganaro\\IRFMN\\In progress\\Helma - aggiornamento modelli Nestlè\\descrittori D5 completi\\helma_desc_vega.txt");
    }
    
}
